# Packages
using DataFrames
using DelimitedFiles
using XLSX
using JuMP
using Gurobi

# Initialization
# A. Reading the required data
classrooms_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "Classrooms"))
programs_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "Programs"))
schedules_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "Schedules"))
general_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "General"))
availability_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "Availability"))
type_table = DataFrame(XLSX.readtable("Thesis/AlgorithmC.xlsx", "Type"))

# Definition of indices
R = classrooms_table[1:21,2] # Classrooms
D = schedules_table[1:5,2]   # Days
T = schedules_table[1:24,1]  # Timeslots
I = general_table[1:22,1]    # Instructors
C = programs_table[1:42,1]   # Courses
S = general_table[1:3,2]     # Student group

# Get the number of corresponding indices
nb_rooms = size(classrooms_table, 1)
nb_days = 5 
nb_timeslots = size(schedules_table, 1)
nb_instructors = size(general_table, 1)
nb_courses = size(programs_table, 1)
nb_students = 3

# Initialize an empty 3D array to store instructors' availability
availability_matrix = zeros(Int, nb_instructors, nb_days, nb_timeslots)

# Populate the availability matrix based on the provided data
for row in eachrow(availability_table)
    timeslot = row.timeslots
    day = row.days
    for (idx, instructor) in enumerate(I)
        availability = row[instructor]
        availability_value = availability == 1 ? 1 : 0
        instructor_idx = findfirst(x -> x == instructor, I)
        day_idx = findfirst(x -> x == day, D)
        timeslot_idx = findfirst(x -> x == timeslot, T)
        availability_matrix[instructor_idx, day_idx, timeslot_idx] = availability_value
    end
end
println(availability_matrix)

# Step 1: Calculate the availability for each instructor
instructor_availability = Dict{String, Int}()
for (idx, instructor) in enumerate(I)
    availability_count = sum(availability_matrix[idx, :, :] .== 1)
    instructor_availability[instructor] = availability_count
end
println(instructor_availability)

# Step 2: Associate each course with its respective instructor
course_instructor = Dict{String, String}()
for row in eachrow(programs_table)
    course_instructor[row.Courses] = row.Instructors
end
println(course_instructor)

# Step 3: Sort the courses list by the availability of the instructor
sorted_courses = sort(collect(keys(course_instructor)), by=x->(instructor_availability[course_instructor[x]], x))

# Step 4: Sort the courses list further by the order of student groups
student_group_order = Dict{String, Int}()
for (idx, group) in enumerate(S)
    student_group_order[group] = idx
end
sorted_courses = sort(sorted_courses, by=x->student_group_order[programs_table[findfirst(isequal(x), programs_table[:, "Courses"]), "Students"]])

# Define the unavailable timeslots
unavailable_timeslots = [1, 2, 9, 10, 21, 22, 23, 24]

# Initialize an empty matrix for course availability
course_availability_matrix = zeros(Int, nb_courses, nb_days, nb_timeslots)

# Populate the course_availability_matrix with availability data for each course
for (course_idx, course) in enumerate(sorted_courses)
    instructor = programs_table[programs_table[:, "Courses"] .== course, "Instructors"][1]
    instructor_index = findfirst(I .== instructor)
    course_availability_matrix[course_idx, :, :] = availability_matrix[instructor_index, :, :]
end

# Update the course_availability_matrix to mark unavailable timeslots as 0
for day_idx in 1:nb_days
    for timeslot in unavailable_timeslots
        course_availability_matrix[:, day_idx, timeslot] .= 0
    end
end

# Create a vector of classrooms sorted by capacity
sorted_classroom_indices = sortperm(classrooms_table[:, "Capacity"])
sorted_classrooms = classrooms_table[sorted_classroom_indices, :]

# Initialize the room availability matrix in the same order as sorted_classrooms
room_availability_matrix = ones(Int, nb_rooms, nb_days, nb_timeslots)
room_availability_matrix = room_availability_matrix[sorted_classroom_indices, :, :]


# Initialize lists to keep track of courses
unscheduled_courses = []
noroom_courses = []
scheduled_courses = Dict{String, Tuple{Int, Int, Int, String}}()


# Iterate over each course in the sorted list
for (course_idx, course) in enumerate(sorted_courses)
    
    course_duration = programs_table[programs_table[:, "Courses"] .== course, "Duration_in_timeslots"][1]
    current_course_students = programs_table[programs_table[:, "Courses"] .== course, "Students"][1]  # Use the "Students" column

    # Initialize feasibility flags
    found_feasible_slot = false
    found_suitable_classroom = false
    
    # Find the feasible time slots for this course
    for day_idx in 1:nb_days
        for start_slot in 1:(nb_timeslots - course_duration + 1)
            # Check if all time slots for the duration of the course are available
            if all(course_availability_matrix[course_idx, day_idx, start_slot:(start_slot + course_duration - 1)] .== 1)
                # Attempt to find a suitable classroom
                for room_idx in 1:nb_rooms
                    # Check if the room is available for the entire duration of the course
                    if all(room_availability_matrix[room_idx, day_idx, start_slot:(start_slot + course_duration - 1)] .== 1)
                        room_capacity = sorted_classrooms[room_idx, "Capacity"]
                        if room_capacity >= programs_table[programs_table[:, "Courses"] .== course, "number_of_students"][1]
                            # Update room availability matrix
                            for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                room_availability_matrix[room_idx, day_idx, timeslot_idx] = 0
                            end
                            # Store the scheduled time slots and assigned classroom for this course
                            scheduled_courses[course] = (day_idx, start_slot, start_slot + course_duration - 1, sorted_classrooms[room_idx, "Classroom"])
                            # Update course availability matrix
                            for other_course_idx in 1:length(sorted_courses)
                                other_course_students = programs_table[programs_table[:, "Courses"] .== sorted_courses[other_course_idx], "Students"][1]
                                if other_course_students == current_course_students && other_course_idx != course_idx
                                    for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                    end
                                end
                            end

                                # Update course availability matrix for courses with the same instructor
                                instructor = programs_table[programs_table[:, "Courses"] .== course, "Instructors"][1]
                                for other_course_idx in 1:length(sorted_courses)
                                    other_course = sorted_courses[other_course_idx]
                                    other_course_instructor = programs_table[programs_table[:, "Courses"] .== other_course, "Instructors"][1]
                                    if other_course_instructor == instructor && other_course != course
                                        for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                            course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                        end
                                    end
                                end

                            # Update course availability matrix based on Type Table
                            for row in eachrow(type_table)
                                if course in row[1:2]
                                    course_idx_in_type = findfirst(x -> x == course, sorted_courses)
                                    other_course = filter(x -> x != course, collect(row[1:2]))
                                    other_course_idx_in_type = findall(x -> x in other_course, sorted_courses)
                                    
                                    for other_course_idx in other_course_idx_in_type
                                        if other_course_idx != course_idx_in_type
                                            for timeslot in 1:nb_timeslots
                                                course_availability_matrix[other_course_idx, day_idx, timeslot] = 0
                                            end
                                        end
                                    end
                                end
                            end
                            
                            found_suitable_classroom = true
                            break
                        end
                    end
                end
                
                if found_suitable_classroom
                    found_feasible_slot = true
                    break
                end
            end
        end
        if found_feasible_slot
            break
        end
    end
    
    # If no feasible time slot is found, put the course in unscheduled course vector
    if !found_feasible_slot
        push!(unscheduled_courses, course)
    end
end


println(scheduled_courses)
println(unscheduled_courses)



# Function to reset time slots that were unavailable
function reset_time_slots(course_availability_matrix, type_table, unscheduled_courses, sorted_courses)
    # Time slots to reset
    reset_slots = [1, 2, 21, 22, 23, 24]
    nb_days = size(course_availability_matrix, 2)
    
    # For every courses
    for course_idx in 1:size(course_availability_matrix, 1)
        for day_idx in 1:nb_days
            for slot in reset_slots
                course_availability_matrix[course_idx, day_idx, slot] = 1
            end
        end
    end
    
    # Apply Type Table constraint
    for row in eachrow(type_table)
        course1, course2 = row[1], row[2]
        course1_idx = findfirst(x -> x == course1, sorted_courses)
        course2_idx = findfirst(x -> x == course2, sorted_courses)
        if course1_idx !== nothing && course2_idx !== nothing
            for day_idx in 1:nb_days
                # For Each day, update the matrix
                if any(course_availability_matrix[course1_idx, day_idx, :] .== 0)
                    for slot in reset_slots
                        course_availability_matrix[course2_idx, day_idx, slot] = 0
                    end
                end
                if any(course_availability_matrix[course2_idx, day_idx, :] .== 0)
                    for slot in reset_slots
                        course_availability_matrix[course1_idx, day_idx, slot] = 0
                    end
                end
            end
        end
    end
end

# Use of the function
reset_time_slots(course_availability_matrix, type_table, unscheduled_courses, sorted_courses)
# Iterates over each unscheduled courses
for unscheduled_course in unscheduled_courses
    course_duration = programs_table[programs_table[:, "Courses"] .== unscheduled_course, "Duration_in_timeslots"][1]
    current_course_students = programs_table[programs_table[:, "Courses"] .== unscheduled_course, "Students"][1]

    # Find feasible time slots for the course
    found_feasible_slot = false
    found_suitable_classroom = false
    
    for day_idx in 1:nb_days
        for start_slot in 1:(nb_timeslots - course_duration + 1)
            if all(course_availability_matrix[findfirst(x -> x == unscheduled_course, sorted_courses), day_idx, start_slot:(start_slot + course_duration - 1)] .== 1)
                for room_idx in 1:nb_rooms
                    if all(room_availability_matrix[room_idx, day_idx, start_slot:(start_slot + course_duration - 1)] .== 1)
                        room_capacity = sorted_classrooms[room_idx, "Capacity"]
                        if room_capacity >= programs_table[programs_table[:, "Courses"] .== unscheduled_course, "number_of_students"][1]
                            # Update Room availability matrix
                            for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                room_availability_matrix[room_idx, day_idx, timeslot_idx] = 0
                            end
                            
                            # Store time slots and classroom
                            scheduled_courses[unscheduled_course] = (day_idx, start_slot, start_slot + course_duration - 1, sorted_classrooms[room_idx, "Classroom"])
                            
                            # Update the availability matrix for the same group of students
                            for other_course_idx in 1:length(sorted_courses)
                                other_course_students = programs_table[programs_table[:, "Courses"] .== sorted_courses[other_course_idx], "Students"][1]
                                if other_course_students == current_course_students
                                    for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                    end
                                end
                            end
                            
                            # Update course availability matrix for courses with the same instructor
                            instructor = programs_table[programs_table[:, "Courses"] .== unscheduled_course, "Instructors"][1]
                            for other_course_idx in 1:length(sorted_courses)
                                other_course = sorted_courses[other_course_idx]
                                other_course_instructor = programs_table[programs_table[:, "Courses"] .== other_course, "Instructors"][1]
                                if other_course_instructor == instructor && other_course != unscheduled_course
                                    for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                    end
                                end
                            end


                            # Update availability matrix based on Type Table
                            for row in eachrow(type_table)
                                if unscheduled_course in row[1:2]
                                    course_idx_in_type = findfirst(x -> x == unscheduled_course, sorted_courses)
                                    other_course = filter(x -> x != unscheduled_course, collect(row[1:2]))
                                    other_course_idx_in_type = findall(x -> x in other_course, sorted_courses)
                                    
                                    for other_course_idx in other_course_idx_in_type
                                        if other_course_idx != course_idx_in_type
                                            for timeslot in 1:nb_timeslots
                                                course_availability_matrix[other_course_idx, day_idx, timeslot] = 0
                                            end
                                        end
                                    end
                                end
                            end
                            
                            found_suitable_classroom = true
                            break
                        end
                    end
                end
                
                if found_suitable_classroom
                    found_feasible_slot = true
                    break
                end
            end
        end
        if found_feasible_slot
            break
        end
    end
    
    # If no feasible time slots are found => new loop to move some courses to other time slots to make the classroom available
    if !found_feasible_slot
        for day_idx in 1:nb_days
            for start_slot in 1:(nb_timeslots - course_duration + 1)
                if all(course_availability_matrix[findfirst(x -> x == unscheduled_course, sorted_courses), day_idx, start_slot:(start_slot + course_duration - 1)] .== 1)
                    for room_idx in 1:nb_rooms
                        occupied_courses = []
                        room_capacity = sorted_classrooms[room_idx, "Capacity"]
                        if room_capacity >= programs_table[programs_table[:, "Courses"] .== unscheduled_course, "number_of_students"][1]
        
                            # Find the courses that use the classroom
                            for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                occupied_course_idx = findfirst(x -> room_availability_matrix[room_idx, day_idx, timeslot_idx] .== 0, 1:length(sorted_courses))
                                if occupied_course_idx !== nothing
                                    push!(occupied_courses, sorted_courses[occupied_course_idx])
                                end
                            end
                            
                            # Move the course.s to other time slots
                            can_move_courses = true
                            for occupied_course in occupied_courses
                                occupied_course_duration = programs_table[programs_table[:, "Courses"] .== occupied_course, "Duration_in_timeslots"][1]
                                found_alternative_slot = false
                                
                                for alt_day_idx in 1:nb_days
                                    for alt_start_slot in 1:(nb_timeslots - occupied_course_duration + 1)
                                        if all(course_availability_matrix[findfirst(x -> x == occupied_course, sorted_courses), alt_day_idx, alt_start_slot:(alt_start_slot + occupied_course_duration - 1)] .== 1) &&
                                        all(room_availability_matrix[room_idx, alt_day_idx, alt_start_slot:(alt_start_slot + occupied_course_duration - 1)] .== 1)
                                            # Update Room Availability Matrix
                                            for timeslot_idx in alt_start_slot:(alt_start_slot + occupied_course_duration - 1)
                                                room_availability_matrix[room_idx, alt_day_idx, timeslot_idx] = 0
                                            end
                                            for timeslot_idx in start_slot:(start_slot + occupied_course_duration - 1)
                                                room_availability_matrix[room_idx, day_idx, timeslot_idx] = 1
                                            end
                                            # Update course Availability matrix for the same student group
                                            for other_course_idx in 1:length(sorted_courses)
                                                other_course_students = programs_table[programs_table[:, "Courses"] .== sorted_courses[other_course_idx], "Students"][1]
                                                if other_course_students == current_course_students
                                                    for timeslot_idx in alt_start_slot:(alt_start_slot + occupied_course_duration - 1)
                                                        course_availability_matrix[other_course_idx, alt_day_idx, timeslot_idx] = 0
                                                    end
                                                    for timeslot_idx in start_slot:(start_slot + occupied_course_duration - 1)
                                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 1
                                                    end
                                                end
                                            end

                                            # Update course availability matrix for the same instructor
                                            instructor = programs_table[programs_table[:, "Courses"] .== occupied_course, "Instructors"][1]
                                            for other_course_idx in 1:length(sorted_courses)
                                                other_course = sorted_courses[other_course_idx]
                                                other_course_instructor = programs_table[programs_table[:, "Courses"] .== other_course, "Instructors"][1]
                                                if other_course_instructor == instructor && other_course != occupied_course
                                                    for timeslot_idx in alt_start_slot:(alt_start_slot + occupied_course_duration - 1)
                                                        course_availability_matrix[other_course_idx, alt_day_idx, timeslot_idx] = 0
                                                    end
                                                    for timeslot_idx in start_slot:(start_slot + occupied_course_duration - 1)
                                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 1
                                                    end
                                                end
                                            end

                                            # Update course availability matrix based on Type Table
                                            for row in eachrow(type_table)
                                                if occupied_course in row[1:2]
                                                    course_idx_in_type = findfirst(x -> x == occupied_course, sorted_courses)
                                                    other_course = filter(x -> x != occupied_course, collect(row[1:2]))
                                                    other_course_idx_in_type = findall(x -> x in other_course, sorted_courses)
                                                    
                                                    for other_course_idx in other_course_idx_in_type
                                                        if other_course_idx != course_idx_in_type
                                                            for timeslot_idx in alt_start_slot:(alt_start_slot + occupied_course_duration - 1)
                                                                course_availability_matrix[other_course_idx, alt_day_idx, timeslot_idx] = 0
                                                            end
                                                            for timeslot_idx in start_slot:(start_slot + occupied_course_duration - 1)
                                                                course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 1
                                                            end
                                                        end
                                                    end
                                                end
                                            end
                                            
                                            found_alternative_slot = true
                                            break
                                        end
                                    end
                                    if found_alternative_slot
                                        break
                                    end
                                end
                                
                                if !found_alternative_slot
                                    can_move_courses = false
                                    break
                                end
                            end
                        end
                        if can_move_courses
                            # Update room availability matrix
                            for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                room_availability_matrix[room_idx, day_idx, timeslot_idx] = 0
                            end
                            
                            # Store new scheduled courses
                            scheduled_courses[unscheduled_course] = (day_idx, start_slot, start_slot + course_duration - 1, classrooms_table[room_idx, "Classroom"])
                            
                            # Update course availability matrix for the course having the same student group
                            for other_course_idx in 1:length(sorted_courses)
                                other_course_students = programs_table[programs_table[:, "Courses"] .== sorted_courses[other_course_idx], "Students"][1]
                                if other_course_students == current_course_students
                                    for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                    end
                                end
                            end
                            
                            # Update course availability matrix for courses with the same instructor
                            instructor = programs_table[programs_table[:, "Courses"] .== unscheduled_course, "Instructors"][1]
                            for other_course_idx in 1:length(sorted_courses)
                                other_course = sorted_courses[other_course_idx]
                                other_course_instructor = programs_table[programs_table[:, "Courses"] .== other_course, "Instructors"][1]
                                if other_course_instructor == instructor && other_course != unscheduled_course
                                    for timeslot_idx in start_slot:(start_slot + course_duration - 1)
                                        course_availability_matrix[other_course_idx, day_idx, timeslot_idx] = 0
                                    end
                                end
                            end

                            # update course availability matrix based on type table
                            for row in eachrow(type_table)
                                if unscheduled_course in row[1:2]
                                    course_idx_in_type = findfirst(x -> x == unscheduled_course, sorted_courses)
                                    other_course = filter(x -> x != unscheduled_course, collect(row[1:2]))
                                    other_course_idx_in_type = findall(x -> x in other_course, sorted_courses)
                                    
                                    for other_course_idx in other_course_idx_in_type
                                        if other_course_idx != course_idx_in_type
                                            for timeslot in start_slot:(start_slot + course_duration - 1)
                                                course_availability_matrix[other_course_idx, day_idx, timeslot] = 0
                                            end
                                        end
                                    end
                                end
                            end

                            found_feasible_slot = true
                            break
                        end
                    end
                end
            end
            if found_feasible_slot
                break
            end
        end
    end
end

# Define unscheduled courses
unscheduled_courses = setdiff(sorted_courses, keys(scheduled_courses))

println("Scheduled courses: ", scheduled_courses)
println("Unscheduled courses: ", unscheduled_courses)
